package simFil;
/*
 * Bug.java... a rod shaped bacteria with cylindrical midsection and hemispherical caps
 *
 */

/*  <ParMSpindle - a simulation of ParM based plasmid segregation>
    Copyright (C) <2008>  <Jonathan B. Alberts>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.awt.*;

import javax.swing.*;
import java.lang.Math.*;
import java.text.DecimalFormat;

import com.sun.j3d.utils.geometry.*;
import com.sun.j3d.utils.universe.*;
import javax.media.j3d.*;
import javax.vecmath.*;

public class Bug extends Thing {
	static double length;			// length from end to end
	static double radius;			// radius of spherical caps and cylindrical section
	static double cellThickness = 0.00;		// thickness of the bacterial cell wall
	static double cylLength;
	static double halfCylLength;
	static double cylVolume;		// volume of the cylinder
	static double sphVolume;		// volume of the 2 hemispheres
	static double bugVolume;		// entire volume contained within the rod-shaped bug
	static double microMolarChangePerMonomer;	// how much to change concentration when one monomer is used or returned
	static Color bugColor = Color.white;
	static float initialTrans = 0.6f;	// initial transparency
	static float initialEmit = 0.1f;	// initial emmissive color
	
	static DecimalFormat timeFormat = new DecimalFormat ("0000.0000");
	static DecimalFormat concFormat = new DecimalFormat ("0.0000");
	
	// for graphics
	static BranchGroup bugG,infoG;
	static Cylinder myCyl1,myCyl2;
	static SphereSection_ lArch,rArch,lArchBony,rArchBony;
	static Color3f ambientC = new Color3f(0.0f,0.0f,0.0f);
	static Color3f diffuseC = new Color3f(0.9f,0.9f,0.9f);
	static Color3f specularC = new Color3f(1.0f,1.0f,1.0f);
	static Color3f emissiveC = new Color3f(0.1f,0.1f,0.1f);
	static float shiny = 60.0f;
	static Appearance aWire = new Appearance();
	static Appearance aSphCaps, bonyApp;
	static TransparencyAttributes noSeeMeTA, bonyTA, tA;
	static Text3D timeTextGeom,concTextGeom;
	static boolean bugInScene = false;
	static boolean coordSysInScene = false;
	static boolean appearanceChanged = false;
//	 for coord sys drawing
	static Color3f xLineColor = new Color3f(1.0f,0.0f,0.0f);
	static Color3f yLineColor = new Color3f(0.0f,1.0f,0.0f);
	static Color3f zLineColor = new Color3f(0.0f,0.0f,1.0f);
	static Appearance xLineApp = new Appearance();
	static Appearance yLineApp = new Appearance();
	static Appearance zLineApp = new Appearance();
	static ColoringAttributes xLineCA = new ColoringAttributes(xLineColor, ColoringAttributes.FASTEST);
	static ColoringAttributes yLineCA = new ColoringAttributes(yLineColor, ColoringAttributes.FASTEST);
	static ColoringAttributes zLineCA = new ColoringAttributes(zLineColor, ColoringAttributes.FASTEST);
	LineArray xLine, yLine, zLine;
	Shape3D xLineShape, yLineShape, zLineShape;
	Pt3D xLineEndPt = new Pt3D();
	Pt3D yLineEndPt = new Pt3D();
	Pt3D zLineEndPt = new Pt3D();
	
	// Pt3Ds reused in collision calculations
	Pt3D forceVec = new Pt3D();
	Pt3D hemDist = new Pt3D();
	Pt3D objVec = new Pt3D();
	Pt3D yzVec = new Pt3D();
	
	public Bug (Pt3D coord, double len, double rad) {
		super (coord);
		length = len;
		radius = rad;
		cylLength = length-2*radius;
		halfCylLength = cylLength/2;
		cylVolume = (length-2*radius)*Math.pow(radius,2)*Math.PI;
		sphVolume = 4*Math.PI*Math.pow(radius,3)/3;
		bugVolume = cylVolume + sphVolume;  // volume in cubic microns
		microMolarChangePerMonomer = (Math.pow(1e5,3)*1e6/(bugVolume*Env.AvogadroNum));		// (1e-6)^4 is for units.
		//talkln ("Bug contains " + String.valueOf(getParMConc()/microMolarChangePerMonomer));
	}
	
	public double getLength() {
		return length;
	}
	
	public double getRadius() {
		return radius;
	}
	

	public static void makeInitialBug () {
		Thing.theBug = new Bug(new Pt3D(0,0,0), Env.bugLength, Env.bugRadius);
	}
	
	public Pt3D rdmPtInsideBug () {
		Pt3D pt = new Pt3D();
		double fracVolInCyl = cylVolume/bugVolume;
		if (Env.mtRNG.raw() < fracVolInCyl) {  //random point in cylinder
			double yC = Env.mtRNG.raw()*2-1;
			double zC = Env.mtRNG.raw()*2-1;
			while ((yC*yC + zC+zC > 1)) { zC = Env.mtRNG.raw()*2-1; }
			pt.y = yC*radius;
			pt.z = zC*radius;
			pt.x = (Env.mtRNG.raw()*2-1)*halfCylLength;
		} else {							// random point in one of the hemispheres
			pt = Pt3D.RandomUnitVec();
			pt.scale(radius*Env.mtRNG.raw());
			if (pt.x > 0) {
				pt.x += halfCylLength;
			} else {
				pt.x += -1*halfCylLength;
			}
		}
		pt.inc(coord);
		return pt;
	}
	
	public Pt3D rdmPtInsideBug (double objRadius) {
		Pt3D pt = new Pt3D();
		double fracVolInCyl = cylVolume/bugVolume;
		if (Env.mtRNG.raw() < fracVolInCyl) {  //random point in cylinder
			double yC = Env.mtRNG.raw()*2-1;
			double zC = Env.mtRNG.raw()*2-1;
			while ((yC*yC + zC*zC > 1)) { zC = Env.mtRNG.raw()*2-1; }
			pt.y = yC*(radius-objRadius);
			pt.z = zC*(radius-objRadius);

			pt.x = (Env.mtRNG.raw()*2-1)*halfCylLength;
		} else {							// random point in one of the hemispheres
			pt = Pt3D.RandomUnitVec();
			pt.scale((radius-objRadius)*Env.mtRNG.raw());
			if (pt.x > 0) {
				pt.x += halfCylLength;
			} else {
				pt.x += -1*halfCylLength;
			}
		}
		pt.inc(coord);
		return pt;
	}
	
	public CollisionEvent amIcolliding (Pt3D ctr, double R) {
		CollisionEvent CE = null;
		double dist = Pt3D.ptDist(ctr, coord);
		objVec.sub(ctr, coord);
		yzVec.copy(0, objVec.y, objVec.z);
		double yzDist = Pt3D.vecMag(yzVec);
		if (Math.abs(objVec.x) < halfCylLength) {	// check collisions in cylindrical section of bug
			if (yzDist+R < radius) {
				return CE;
			} else {
				double impingedist = Math.abs(yzDist + R - radius);
				yzVec.unitVec();
				yzVec.reverse();  // yzVec is useful as a force vector
				CE = new CollisionEvent(impingedist, yzVec);
				return CE;
			 }
			
		} else {		// check collisions in hemispherical caps
			if (objVec.x > 0) {	// positive x hemisphere
				hemDist.copy(coord.x+halfCylLength, coord.y, coord.z);
				double topdist = Pt3D.ptDist (ctr, hemDist);
				if (topdist + R < radius) {
					return CE;
				} else {
					double impingedist = Math.abs(topdist + R - radius);
					forceVec.unitVec(topdist, hemDist, ctr);
					CE = new CollisionEvent(impingedist, forceVec);
					return CE;
				}
			}
			if (objVec.x < 0) {  // negative x hemisphere
				hemDist.copy(coord.x-halfCylLength, coord.y, coord.z);
				double botdist = Pt3D.ptDist (ctr, hemDist);
				if (botdist + R < radius) {
					return CE;
				} else {
					double impingedist = Math.abs(botdist + R - radius);
					forceVec.unitVec(botdist, hemDist, ctr);
					CE = new CollisionEvent(impingedist, forceVec);
					return CE;
				}
			}
		
		}
		return CE;
	}
	
	public static void appearanceChange() {
		appearanceChanged = true;
		if (Env.paused) { theBug.updateGraphics();}
	}
	
	public void setWireFrameAppearance () {
		a.setLineAttributes(new LineAttributes(0.5f, LineAttributes.PATTERN_DASH,true));
		aSphCaps.setLineAttributes(new LineAttributes(0.5f, LineAttributes.PATTERN_DASH,true));;
		PolygonAttributes wireAtts = new PolygonAttributes(PolygonAttributes.POLYGON_LINE,PolygonAttributes.CULL_FRONT,0.0f,true);
		a.setPolygonAttributes(wireAtts);
		PolygonAttributes wireSphCapAtts = new PolygonAttributes(PolygonAttributes.POLYGON_LINE,PolygonAttributes.CULL_BACK,0.0f,true);
		aSphCaps.setPolygonAttributes(wireSphCapAtts);
	}
	
	public void makeGraphics () {
		setGraphicsCapabilities();	// set capabilities for BranchGroup, TransformGroup, and Appearance
		// no see um TransparencyAttributes
		noSeeMeTA = new TransparencyAttributes(TransparencyAttributes.NICEST,1.0f);
		
		// the main appearance
		m = new Material (ambientC,emissiveC,diffuseC,specularC,shiny);
		m.setCapability(Material.ALLOW_COMPONENT_WRITE);
		ColoringAttributes cA = new ColoringAttributes(1.0f,1.0f,1.0f,ColoringAttributes.NICEST);
		tA = new TransparencyAttributes (TransparencyAttributes.NICEST,Bug.initialTrans );
		tA.setCapability(TransparencyAttributes.ALLOW_VALUE_WRITE);
		a.setMaterial(m);
		a.setColoringAttributes(cA);
		a.setTransparencyAttributes(tA);
		
		// make alternate transparent appearance for cylinder end caps
		Appearance capA = new Appearance();
		ColoringAttributes capCA = new ColoringAttributes(0.0f,0.0f,0.0f,ColoringAttributes.NICEST);
		capA.setColoringAttributes(capCA);
		capA.setTransparencyAttributes(noSeeMeTA);
		capA.setLineAttributes(new LineAttributes(0.5f, LineAttributes.PATTERN_DASH_DOT,true));
		capA.setPolygonAttributes(new PolygonAttributes(PolygonAttributes.POLYGON_POINT,PolygonAttributes.CULL_BACK,0.0f));
		
		// make another appearance for spherical ends
		aSphCaps = new Appearance ();
		aSphCaps.setCapability(Appearance.ALLOW_POLYGON_ATTRIBUTES_WRITE);
		aSphCaps.setCapability(Appearance.ALLOW_LINE_ATTRIBUTES_WRITE);
		aSphCaps.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_WRITE);
		aSphCaps.setMaterial(m);
		aSphCaps.setColoringAttributes(cA);
		aSphCaps.setTransparencyAttributes(tA);
		
		// make another appearance for second cylinder
		bonyApp = new Appearance ();
		bonyApp.setCapability(Appearance.ALLOW_POLYGON_ATTRIBUTES_WRITE);
		bonyApp.setCapability(Appearance.ALLOW_LINE_ATTRIBUTES_WRITE);
		bonyApp.setCapability(Appearance.ALLOW_TRANSPARENCY_ATTRIBUTES_WRITE);
		bonyApp.setMaterial(m);
		bonyApp.setColoringAttributes(cA);
		bonyTA = new TransparencyAttributes(TransparencyAttributes.NICEST,0.9f);
		bonyApp.setTransparencyAttributes(bonyTA);
		bonyApp.setLineAttributes(new LineAttributes(1.0f, LineAttributes.PATTERN_SOLID,true));
		PolygonAttributes bonyAtts = new PolygonAttributes(PolygonAttributes.POLYGON_LINE,PolygonAttributes.CULL_NONE,0.0f);
		bonyAtts.setBackFaceNormalFlip(true);
		bonyApp.setPolygonAttributes(bonyAtts);
		
		setWireFrameAppearance ();  // start out with the wireframe appearance
		
		//build the bug shape... first the cylindrical section.. 2 cylinders
		myCyl1 = new Cylinder((float)radius,(float)(cylLength),Primitive.GENERATE_NORMALS,16,16,a);
		myCyl1.setCapability(myCyl1.ENABLE_APPEARANCE_MODIFY);
		myCyl1.setAppearance(myCyl1.TOP,capA);
		myCyl1.setAppearance(myCyl1.BOTTOM,capA);
		Transform3D cylT3D = new Transform3D();
		cylT3D.rotZ(Math.PI/2);
		TransformGroup cylTG = new TransformGroup(cylT3D);
		cylTG.addChild(myCyl1);
		
		myCyl2 = new Cylinder((float)(radius+cellThickness),(float)cylLength,Primitive.GENERATE_NORMALS,16,16,bonyApp);
		myCyl2.setCapability(myCyl2.ENABLE_APPEARANCE_MODIFY);
		myCyl2.setAppearance(myCyl1.TOP,capA);
		myCyl2.setAppearance(myCyl1.BOTTOM,capA);
		Transform3D cyl2T3D = new Transform3D();
		cyl2T3D.rotZ(Math.PI/2);
		TransformGroup cyl2TG = new TransformGroup(cyl2T3D);
		cyl2TG.addChild(myCyl2);
		
		// right hemispherical cap (arch)
		rArch = new SphereSection_ (0.0, Math.PI/2, 12, 2*Math.PI, 0.0, 17, radius, aSphCaps);
		rArch.setCapability(rArch.ALLOW_APPEARANCE_WRITE);
		Transform3D rArchT3D = new Transform3D();
		rArchT3D.rotZ(-Math.PI/2);
		Transform3D tweakRot = new Transform3D();
		tweakRot.rotY(6*Math.PI/180);
		//rArchT3D.mul(tweakRot);
		rArchT3D.setTranslation(new Vector3d(cylLength/2,0,0));
		TransformGroup rArchTG = new TransformGroup(rArchT3D);
		
		// bony right hemispherical cap (arch)
		rArchBony = new SphereSection_ (0.0, Math.PI/2, 12, 2*Math.PI, 0.0, 17, radius, bonyApp);
		rArchBony.setCapability(rArch.ALLOW_APPEARANCE_WRITE);
		rArchTG.addChild(rArch);
		rArchTG.addChild(rArchBony);

		// left hemispherical cap (arch)
		lArch = new SphereSection_ (0.0, Math.PI/2, 12, 2*Math.PI, 0.0, 17, radius, aSphCaps);
		lArch.setCapability(lArch.ALLOW_APPEARANCE_WRITE);
		Transform3D lArchT3D = new Transform3D();
		lArchT3D.rotZ(Math.PI/2);
		lArchT3D.setTranslation(new Vector3d(-cylLength/2,0,0));
		TransformGroup lArchTG = new TransformGroup(lArchT3D);
		
		// left hemispherical cap (arch)
		lArchBony = new SphereSection_ (0.0, Math.PI/2, 12, 2*Math.PI, 0.0, 17, radius, bonyApp);
		lArchBony.setCapability(lArch.ALLOW_APPEARANCE_WRITE);
		lArchTG.addChild(lArch);
		lArchTG.addChild(lArchBony);
		
		g3d.setTransform(t3d);
		
		g3d.addChild(cylTG);
		g3d.addChild(cyl2TG);
		g3d.addChild(lArchTG);
		g3d.addChild(rArchTG);
		
		bugG = new BranchGroup();
		bugG.setCapability(BranchGroup.ALLOW_DETACH);
		bugG.addChild(g3d);

		if (Env.bugOn) {	
			G.addChild(bugG);
			bugInScene = true;
		}
		
		G.compile();
		graphicsMade = true;
	}
	
	public void updateGraphics () {
		if (!Env.bugOn) { 
			if (bugInScene) {
				bugG.detach();
				bugInScene = false;
			}
		} else {
			if (!bugInScene) {
				G.addChild(bugG);
				bugInScene = true;
			}
		}
		
	}
}
