package simFil;
/*
 * Plasmid.java
 */

/*  <SimFil - simulation of cytoskeletal filament flexibility by Pairwise Agent Interaction with Rational Superposition (PAIRS)>
    Copyright (C) <2008>  <Jonathan B. Alberts>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


import java.awt.*;
import javax.swing.*;
import java.lang.Math.*;
import com.sun.j3d.utils.geometry.Sphere;
import com.sun.j3d.utils.geometry.Box;
import com.sun.j3d.utils.universe.*;
import javax.media.j3d.*;
import javax.vecmath.*;

import java.io.File;

public class PSphere extends Thing {
	static PSphere [] theSpheres = new PSphere [500];	// array holding all plasmids
	static int sphereCt = 0;			//
	private int boundFilaments = 0;			// counter for number of bound ParM filaments
	private int registeredFilCt = 0;
	private int [] registeredFilIds = new int [50];
	static int maxParRparCs = 100;
	Pt3D [] parRparCs = new Pt3D [maxParRparCs];
	int parRparCCt = 0;
	Color sphereColor = Color.green;	// color for drawing
	Pt3D aVec = new Pt3D(); //recycled
	
	// custom plasmid properties
	boolean customSphere = false;
	boolean immobile = false;
	double customTransDiff = 0;
	double customRotDiff = 0;
	double customRadius = 0;
	boolean xSlider = false;

	Sphere mySphere;
	Sphere lightSphere;
	Box lightBox;
	BranchGroup centerGroup = new BranchGroup();
	Sphere [] parRparCSpheres = new Sphere [maxParRparCs];
	Vector3d coordVec3d = new Vector3d();
	static Color3f ambientC = new Color3f(0.0f,0.0f,0.0f);
	static Color3f diffuseC = new Color3f(0.0f,0.3f,0.6f);
	static Color3f specularC = new Color3f(1.0f,1.0f,1.0f);
	static Color3f emissiveC = new Color3f(0.0f,0.0f,0.0f);
	
	static Color3f ambientIC = new Color3f(0.0f,0.0f,0.0f);
	static Color3f diffuseIC = new Color3f(0.8f,0.8f,0.8f);
	static Color3f specularIC = new Color3f(1.0f,1.0f,1.0f);
	static Color3f emissiveIC = new Color3f(0.5f,0.8f,1.0f);
	
	static Color3f ambientIM = new Color3f(0.0f,0.0f,0.0f);
	static Color3f diffuseIM = new Color3f(1.0f,1.0f,1.0f);
	static Color3f specularIM = new Color3f(1.0f,1.0f,1.0f);
	static Color3f emissiveIM = new Color3f(1.0f,1.0f,1.0f);
	
	static float shiny = 60.0f;
	static int pLightCt = 2;
	PointLight [] pLights = new PointLight[pLightCt];
	
	public PSphere (Pt3D initCoord) {
		super(initCoord);
		addPlasmid(this);
		calculateProperties();
		initialize();
	}
	
	public void calculateProperties () {
		bTransDiff.x = getTransDiff();
		bTransDiff.y = bTransDiff.x;
		bTransDiff.z = bTransDiff.x;
		if (xSlider) { bTransDiff.x = 1e-12; }
		bTransGam.div(Env.Boltz*Env.tempK,bTransDiff);		// Einstein's relation gamma=kT/D
		
		bRotDiff.x = getRotDiff();
		bRotDiff.y = bRotDiff.x;
		bRotDiff.z = bRotDiff.x;
		bRotGam.div(Env.Boltz*Env.tempK,bRotDiff);		// Einstein's relation gamma=kT/D
		
	}
	
	public void initialize () {
		// this method assumes the unit x and y vectors have been set (though maybe not orthogonal), or are unchanged
		// determine z-unit vectors, then reset y-unit vector to ensure orthogonality with uVec
		zVec.cross(uVec, yVec);
		yVec.cross(zVec, uVec);
		// find the transformation matrices at this time step
		transMat ();
	}
	
	public void incForceSum (Pt3D F, Pt3D vec) {
		incForceSum(F);
	}
	
	public void forceField () {
		//talkln ("in constant F app");
		forceSum.inc(new Pt3D(Env.constantForce.getValue()*1e-12,0,0));
	}
	
	public void linkedForce () {
		// user puts in pN/nM... convert to N/nm and use a dist in nM 
		double k = Env.kAttracted.getValue()*1e-12;	// N/m spring constant
		double impDist = 1e3*Pt3D.ptDist(coord, origin);
		Pt3D fVec = Pt3D.UnitVec(origin,coord);
		fVec.scale(impDist*k);
		forceSum.inc(fVec);
	}
	
	public void moveThing () {
		// Given the forces/torques at this time point... move with explicit Euler approximation to ODE solution
		
		// calculate properties...they don't typically change during a run unless gui-usered
		calculateProperties();
		if (immobile) { return; }
		
		// first check that forceSum and torqueSum aren't wacky... exit method if they are
		if (!forceSum.checkPt3D()) { return; }
		if (!torqueSum.checkPt3D()) { return; }
		
		//talkln ("forceSum for plasmid " + this + " is " + forceSum.reportCoords());
		// Work in coordinates aligned with the plasmid... transform forces and torques into body-fixed axis....
		bForceSum.XTox(this, forceSum);
		bTorqueSum.XTox(this, torqueSum);
		if (!Env.plasmidBrownianMotionOff) {
			// add random forces, given in body-fixed frame
			calcRandomForces();
			// then add to force and torque sums
			bForceSum.inc(randForces);
			bTorqueSum.inc(randTorques);
		}
		
		// now that the forces and torques are in the body fixed frame, we apply the eoms....
		bVeloc.div(1.0e6, bForceSum, bTransGam);		// in micron/sec
		bAngVeloc.div(bTorqueSum, bRotGam);			// in radians/sec
		
		// ** before progressing .... check that bVeloc and bAngVeloc are not NaN... exit if wacky
		if (!bVeloc.checkPt3D()) { talkln ("** problem with bVeloc for filament " + String.valueOf(this)); return; }
		if (!bAngVeloc.checkPt3D()) { talkln ("** problem with bAngVeloc for filament " + String.valueOf(this)); return; }
		
		// ....then transform back into fixed coordinate frame
		veloc.xToX(this, bVeloc);
		coord.inc(Env.deltaT.getValue(), veloc);
		
		// to apply the body-fixed angular velocities, approximate new unit vector from arc of rotations.. good for small rotations
		// for uVec
		double uVecTransInZ = -bAngVeloc.y * Env.deltaT.getValue();	// arclength out at 1 micron
		double uVecTransInY = bAngVeloc.z * Env.deltaT.getValue();
		uVec.setVals(1,uVecTransInY,uVecTransInZ);	// in body-fixed, not a unit vector yet
		uVec.xToX(this);	// make in fixed-frame, not a unit vector yet
		uVec.unitVec();		// make a unit vector
		
		// for yVec 
		double yVecTransInX = - uVecTransInY;
		double yVecTransInZ = bAngVeloc.x * Env.deltaT.getValue();	// arclength at 1 micron
		yVec.setVals(yVecTransInX, 1, yVecTransInZ);
		yVec.xToX(this);
		yVec.unitVec();
		
		initialize();
	}
	
	public void step () {

		if (Env.bugOn) { checkBugCollision(); }
		if (Env.spheresUnderForceField) { forceField(); }
		if (Env.spheresLinked) { linkedForce(); }
		moveThing();
		initialize();
	
		// reset counters etc.
		resetCounters();
	}
	
	public static void checkPlasmidCollision () {
		for (int i=0; i<sphereCt; i++) {
			PSphere iP = theSpheres[i];
			for (int p=i+1;p<sphereCt;p++) {
				PSphere pP = theSpheres[p];
				double pDist = Pt3D.ptDist(iP.coord, pP.coord);
				if (pDist<pP.getRadius()+iP.getRadius()) {
					double impingedist = pP.getRadius()+iP.getRadius() - pDist;
				    Pt3D iVec = Pt3D.UnitVec(pDist, iP.coord, pP.coord);
					Pt3D pVec = Pt3D.Reverse(iVec);
					double mag = (Env.sphCollisionCoeff.getValue()*1.0e-6*impingedist/Env.deltaT.getValue())/(1/iP.bTransGam.x+1/pP.bTransGam.x);
					iP.forceSum.inc(Pt3D.Scale(mag,iVec));
					pP.forceSum.inc(Pt3D.Scale(mag,pVec));
					//Pt3D plasCollForce = Pt3D.Scale(mag, iVec);
					//talkln ("forceSum in plasmid coll. is " + plasCollForce.reportCoords());
					//talkln ("colliding");
				}
			}
		}
	}
	
	public void checkBugCollision () {
		CollisionEvent cE = theBug.amIcolliding(coord,getRadius());
		if (cE == null) { return; }
		// collision!
		double mag = Env.sphCollisionCoeff.getValue()*1.0e-6*cE.delta*bTransGam.x/Env.deltaT.getValue();
		forceSum.inc(Pt3D.Scale(mag,cE.forceUVec));
	}
	
	public void resetCounters() {
		super.resetCounters();	// call the generic Thing method
		registeredFilCt = 0;
	}
	
	public void setCustomPlas (double rad, double transDiff, double rotDiff) {
		customSphere = true;
		customRadius = rad;
		customTransDiff = transDiff;
		customRotDiff = rotDiff;
	}
	
	public void setImmobile () {
		immobile = true;
		customSphere = true;
		customTransDiff = 1e-100;
		customRotDiff = 1e-100;
	}
	
	public double getRadius () {
		if (customSphere) {
			return customRadius;
		} else {
			return Env.sphRadius.getValue();
		}
	}
	
	public double getTransDiff () {
		if (customSphere) { 
			if (customTransDiff > 0) {return customTransDiff; } else { return stokesTransDiff(); }
		}
		if (Env.sphTransDiff.isActive()) { return Env.sphTransDiff.getValue(); }
		return stokesTransDiff();	
	}
	
	public double stokesTransDiff () {
		double stokesLawDrag = 6*Math.PI*Env.aeta.getValue()*getRadius()*1.0e-6; // Stokes law
		return Env.Boltz*Env.tempK/stokesLawDrag;	
	}
	
	public double getRotDiff () {
		if (customSphere) { 
			if (customRotDiff > 0) { return customRotDiff; } else { return stokesRotDiff(); }
		}
		if (Env.sphRotDiff.isActive()) { return Env.sphRotDiff.getValue(); }
		return stokesRotDiff();
	}
	
	public double stokesRotDiff () {
		double stokesLawRotDrag = 8*Math.PI*Env.aeta.getValue()*Math.pow(getRadius()*1.0e-6,3); // drag for turning about x
		return Env.Boltz*Env.tempK/stokesLawRotDrag;
	}
	
	public void filamentOn () {
		boundFilaments ++;
	}
	
	public void filamentOff () {
		boundFilaments --;
		if (boundFilaments < 0) { boundFilaments = 0; }
	}
	
	public void registerFilament (FilSegment boundSeg) {
		registeredFilIds[registeredFilCt] = boundSeg.filID;
		registeredFilCt++;
	}
	
	public boolean filamentBound (int filID) {
		for (int i=0;i<registeredFilCt;i++) {
			if (filID == registeredFilIds[i]) { return true; }
		}
		return false;
	}
	
	public static boolean filamentBoundToAnyPlasmid (int filID) {
		for (int i=0;i<sphereCt;i++) {
			if (theSpheres[i].filamentBound(filID)) { return true; }
		}
		return false;
	}
	
	public static void plasmidsLightSwitch (boolean on) {
		for (int i=0;i<PSphere.sphereCt;i++) {
			for (int j=0;j<pLightCt;j++){
				theSpheres[i].pLights[j].setEnable(on);
			}
		}
	}
	
	public static File cwd() {
		return new File(System.getProperty("user.dir","."));
	}
	 
	public static File resDir() {
		String p = System.getProperty("jexepack.resdir");
		return (p!=null) ? new File(p) : cwd();
	}
	
	public static void switchLightCenters () {
		for (int i=0;i<sphereCt;i++) {
			theSpheres[i].centerGroup.detach();
			theSpheres[i].centerGroup.removeAllChildren();
			theSpheres[i].centerGroup.addChild(theSpheres[i].lightSphere);
			theSpheres[i].g3d.addChild(theSpheres[i].centerGroup);
		}
	}
	
	public void setGraphicsCapabilities() {
		super.setGraphicsCapabilities();
		centerGroup.setCapability(BranchGroup.ALLOW_DETACH);
		centerGroup.setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
	}
	
	public void makeGraphics () {
		// make material
		if (immobile | xSlider) {
			m = new Material(ambientIM,emissiveIM,diffuseIM,specularIM,128);
		} else {
			m = new Material (ambientC,emissiveC,diffuseC,specularC,shiny);
		}
		// set capabilities
		setGraphicsCapabilities();
		
		ColoringAttributes cA = new ColoringAttributes(Env.plasmidColor3f, ColoringAttributes.NICEST);
		a.setColoringAttributes(cA);
		PolygonAttributes plasAtts = new PolygonAttributes(PolygonAttributes.POLYGON_FILL,PolygonAttributes.CULL_FRONT,0.0f);
		plasAtts.setBackFaceNormalFlip(true);
		if (!immobile & !xSlider) {
			TransparencyAttributes tA = new TransparencyAttributes ();
			tA.setTransparency(0.2f);
			tA.setTransparencyMode(tA.NICEST);
			a.setTransparencyAttributes(tA);

			a.setPolygonAttributes(plasAtts);
		}
		a.setMaterial(m);
		mySphere = new Sphere((float)getRadius(),Sphere.GENERATE_NORMALS, Env.plasmidTessalation, a);
		mySphere.setCapability(Sphere.ALLOW_BOUNDS_READ);
		
		Material iMat = new Material(ambientIC,emissiveIC,diffuseIC,specularIC,shiny);
		Appearance iA = new Appearance();
		PolygonAttributes iAAtts = new PolygonAttributes(PolygonAttributes.POLYGON_FILL,PolygonAttributes.CULL_BACK,0.0f);
		plasAtts.setBackFaceNormalFlip(true);
		iA.setPolygonAttributes(iAAtts);
		iA.setMaterial(iMat);    
		//iA.setTransparencyAttributes(tA);
		float lightRad = (float)(getRadius()/3);
		lightSphere = new Sphere(lightRad,Sphere.GENERATE_NORMALS, 12, iA);
		lightBox = new Box(lightRad,lightRad,lightRad,Box.GENERATE_TEXTURE_COORDS,iA);
		centerGroup.addChild(lightSphere);
		
		// ParR-ParC spheres
		ColoringAttributes parCA = new ColoringAttributes(Env.parRparCColor3f, ColoringAttributes.NICEST);
		Appearance parRparCApp = new Appearance();
		parRparCApp.setColoringAttributes(cA);
		for (int i=0;i<parRparCSpheres.length;i++) {
			//parRparCSpheres[i] = new Sphere((float)Env.parRparCRadius,Sphere.GENERATE_NORMALS,Env.plasmidTessalation,parRparCApp);
		}
		
		// fiduciary marks
		double markSphereSize = getRadius()/12.0;
		Appearance yApp = new Appearance();
		ColoringAttributes yCA = new ColoringAttributes(new Color3f(Color.BLACK), ColoringAttributes.NICEST);
		Material yM = new Material (new Color3f(Color.BLACK),emissiveC,new Color3f(Color.BLACK),specularC,shiny);
		yApp.setColoringAttributes(yCA);
		yApp.setMaterial(yM);
		Sphere yMark = new Sphere((float)(markSphereSize),Sphere.GENERATE_NORMALS, Env.plasmidTessalation, yApp);
		Transform3D yMarkT3D = new Transform3D();
		yMarkT3D.setScale(new Vector3d(1.0,0.1,1.0));
		yMarkT3D.setTranslation(new Vector3d(0.0,getRadius(),0.0));
		TransformGroup yMarkTG = new TransformGroup();
		yMarkTG.addChild(yMark);
		yMarkTG.setTransform(yMarkT3D);
		
		Sphere xMark = new Sphere((float)(markSphereSize),Sphere.GENERATE_NORMALS, Env.plasmidTessalation, yApp);
		Transform3D xMarkT3D = new Transform3D();
		xMarkT3D.setScale(new Vector3d(0.1,1.0,1.0));
		xMarkT3D.setTranslation(new Vector3d(getRadius(),0.0,0.0));
		TransformGroup xMarkTG = new TransformGroup();
		xMarkTG.addChild(xMark);
		xMarkTG.setTransform(xMarkT3D);
		
		Appearance textAppear = new Appearance();
	    ColoringAttributes textColor = new ColoringAttributes();
	    textColor.setColor(1.0f, 0.0f, 0.0f);
	    textAppear.setColoringAttributes(textColor);
	    textAppear.setMaterial(new Material());
	    // Create a simple shape leaf node, add it to the scene graph.
	    Font3D font3D = new Font3D(new Font("Helvetica", Font.PLAIN, 1),new FontExtrusion());
	    Text3D yTextGeom = new Text3D(font3D, new String("Y"));
	    yTextGeom.setAlignment(Text3D.ALIGN_CENTER);
	    Shape3D yTextShape = new Shape3D();
	    yTextShape.setGeometry(yTextGeom);
	    yTextShape.setAppearance(textAppear);
	    Transform3D yTextT3D = new Transform3D();
	    TransformGroup yTextTG = new TransformGroup();
	    yTextT3D.rotX(Math.PI/2.0);
	    yTextT3D.setScale(0.015);
	    yTextT3D.setTranslation(new Vector3d(0,.010,-0.005));
	    yTextTG.setTransform(yTextT3D);
	    yTextTG.addChild(yTextShape);
	    yMarkTG.addChild(yTextTG);
		
	    Text3D xTextGeom = new Text3D(font3D, new String("X"));
	    xTextGeom.setAlignment(Text3D.ALIGN_CENTER);
	    Shape3D xTextShape = new Shape3D();
	    xTextShape.setGeometry(xTextGeom);
	    xTextShape.setAppearance(textAppear);
	    Transform3D xTextT3D = new Transform3D();
	    TransformGroup xTextTG = new TransformGroup();
	    xTextT3D.rotY(Math.PI/2.0);
	    xTextT3D.setScale(0.015);
	    xTextT3D.setTranslation(new Vector3d(0.007,-.005,0));

	    xTextTG.setTransform(xTextT3D);
	    xTextTG.addChild(xTextShape);
	    xMarkTG.addChild(xTextTG);
	    
	    // point light
	    for (int i=0;i<pLightCt;i++){
		    pLights[i] = new PointLight(Env.plasmidLightColor3f,new Point3f(0f,0f,0f),new Point3f(1f,1f,1f));
		    pLights[i].setCapability(PointLight.ALLOW_STATE_WRITE);
		    pLights[i].setCapability(PointLight.ALLOW_POSITION_WRITE);
		    pLights[i].setCapability(Light.ALLOW_SCOPE_WRITE);
		    pLights[i].removeAllScopes();
		   // pLights[i].addScope(Bug.bugG);
		    BoundingSphere plasLightBounds = new BoundingSphere(new Point3d( 0.0, 0.0, 0.0 ),3.0 );
		    pLights[i].setInfluencingBounds(plasLightBounds);
		    pLights[i].setAttenuation(0f,20f,0f);
	    }
	    float radMult = 1.1f;
	    float radiusf = (float)getRadius();
	    pLights[0].setPosition(radMult*radiusf,0,0);
	    pLights[1].setPosition(-radMult*radiusf,0,0);
	   // pLights[3].setPosition(0,-radMult*radiusf,0);
	   // pLights[4].setPosition(0,0,radMult*radiusf);
	   // pLights[5].setPosition(0,0,-radMult*radiusf);
		
		g3d.addChild(mySphere);
		g3d.addChild(centerGroup);
		//g3d.addChild(xMarkTG);
		//g3d.addChild(yMarkTG);
		for (int i=0;i<pLightCt;i++) { g3d.addChild(pLights[i]); }
		G.addChild(g3d);
		
		graphicsMade = true;
	}
	public void updateGraphics () {
		coord.copyToVector3d(coordVec3d);
		t3d.setTranslation(coordVec3d);
		t3d.setRotation(mxToX);
		g3d.setTransform(t3d);
	 }
	
	public static void makeTestPlasmidPair () {
		Pt3D loc1 = new Pt3D (Env.tstPlasmidOffset,0,0);
		Pt3D loc2 = new Pt3D (-Env.tstPlasmidOffset,0,0);
		new PSphere(loc1);
		new PSphere(loc2);
	}
	
	public static void addPlasmid (PSphere newPlasmid) {
		theSpheres[sphereCt] = newPlasmid;
		sphereCt ++;
	}
	
	public static void removeAll () {
		for (int i=0;i<sphereCt;i++) {
			theSpheres[i].G.detach();
		}
		for (int i=0;i<sphereCt;i++) {
			theSpheres[i].removeMe = true;
			theSpheres[i] = null;
		}
		sphereCt = 0;
	}
}
